@@ -1,13 +1,14 @@
-#define VERSION "1.86"
+#define VERSION "1.87"
 
 /*
  *  units, a program for units conversion
- *  Copyright (C) 1996, 1997, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006
+ *  Copyright (C) 1996, 1997, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
+ *                2007
  *  Free Software Foundation, Inc
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
+ *  the Free Software Foundation; either version 3 of the License, or
  *  (at your option) any later version.
  * 
  *  This program is distributed in the hope that it will be useful,
@@ -44,10 +45,12 @@
 #  define EOFCHAR "D"
 #endif
 
+#define HOMEUNITSFILE ".units.dat"   /* Units file in home directory */
 #define PRIMITIVECHAR '!'	/* Character that marks irreducible units */
 #define COMMENTCHAR '#'         /* Comments marked by this character */
 #define COMMANDCHAR '!'         /* Unit database commands marked with this */
 #define HELPCOMMAND "help"      /* Command to request help at prompt */
+#define SEARCHCOMMAND "search"  /* Command to request text search of units */
 #define UNITMATCH "?"           /* Command to request conformable units */
 #define DEFAULTPAGER "more"     /* Default pager program */
 #define DEFAULTLOCALE "en_US"   /* Default locale */
@@ -66,7 +69,8 @@ int unitcheck = 0;              /* Flag for unit checking, set to 1 for
                                    regular check, set to 2 for verbose check */
 int verbose = 1;                /* Flag for output verbosity */
 int strictconvert = 0;          /* Strict conversion (disables reciprocals) */
-int minusminus = 1;             /* Should '-' character give subtraction */
+int minusminus = 1;             /* Does '-' character give subtraction */
+int oldstar = 0;                /* Does '*' have higher precedence than '/' */
 int oneline = 0;                /* Suppresses the second line of output */
 char *unitsfiles[MAXFILES+1];   /* Null terminated list of units file names */
 char *progname="units";         /* Used in error messages */
@@ -1283,7 +1287,7 @@ completereduce(struct unittype *unit)
 
 
 /* Raise theunit to the specified power.  This function does not fill
-   in NULLUNIT gaps, which could be considred a deficiency. */
+   in NULLUNIT gaps, which could be considered a deficiency. */
 
 int
 expunit(struct unittype *theunit, int  power)
@@ -1957,25 +1961,34 @@ struct namedef {
   char *def;
 };
     
-void addtolist(struct unittype *have, char *rname, char *name, char *def, 
+#define CONFORMABLE 1
+#define TEXTMATCH 2
+
+void 
+addtolist(struct unittype *have, char *searchstring, char *rname, char *name, char *def, 
 	       struct namedef **list, int *listsize, int *maxnamelen, 
-	       int *count)
+	       int *count, int searchtype)
 {
   struct unittype want;
-  int len;
+  int len,keepit;
 
   if (!name) 
     return;
-  initializeunit(&want);
-  parseunit(&want, name,0,0);
-  completereduce(&want);
-  if (!compareunits(have,&want,ignore_dimless)){
+  if (searchtype==CONFORMABLE){
+    initializeunit(&want);
+    parseunit(&want, name,0,0);
+    completereduce(&want);
+    keepit = !compareunits(have,&want,ignore_dimless);
+  } else if (searchtype == TEXTMATCH) {
+    keepit = (strstr(rname,searchstring) != NULL);
+  }
+  if (keepit){
     if (*count==*listsize){
       *listsize += 100;
       *list = (struct namedef *) 
 	realloc(*list,*listsize*sizeof(struct namedef));
       if (!*list){
-	fprintf(stderr, "%s: memory allocation error (tryallunits)\n",
+	fprintf(stderr, "%s: memory allocation error (addtolist)\n",
 		progname);  
 	exit(3);
       }
@@ -1990,7 +2003,8 @@ void addtolist(struct unittype *have, char *rname, char *name, char *def,
     if (len>*maxnamelen)
 	  *maxnamelen = len;
   }
-  freeunit(&want);
+  if (searchtype == CONFORMABLE)
+    freeunit(&want);
 }
 
 
@@ -2012,18 +2026,19 @@ screensize()
 
 
 /* 
-   Cycle through all units and print ones which are conformable with
-   the argument.
+   If have is non-NULL then search through all units and print the ones
+   which are conformable with have.  Otherwise search through all the 
+   units for ones whose names contain the second argument as a substring.
 */
 
 void 
-tryallunits(struct unittype *have)
+tryallunits(struct unittype *have, char *searchstring)
 {
   struct unitlist *uptr;
   struct namedef *list;
   int listsize, maxnamelen, count;
   struct func *funcptr;
-  int i, j;
+  int i, j, searchtype;
   FILE *outfile;
 
   list = (struct namedef *) mymalloc( 100 * sizeof(struct namedef), 
@@ -2032,17 +2047,27 @@ tryallunits(struct unittype *have)
   maxnamelen = 0;
   count = 0;
 
+  if (have)
+    searchtype = CONFORMABLE;
+  else {
+    if (!searchstring)
+      searchstring="";
+    searchtype = TEXTMATCH;
+  }
+
   for(i=0;i<HASHSIZE;i++)
     for (uptr = utab[i]; uptr; uptr = uptr->next)
-      addtolist(have, uptr->name, uptr->name, uptr->value, &list, 
-		&listsize, &maxnamelen, &count);
+      addtolist(have, searchstring, uptr->name, uptr->name, uptr->value, 
+                &list, &listsize, &maxnamelen, &count, searchtype);
   for(funcptr=firstfunc;funcptr;funcptr=funcptr->next){
     if (funcptr->table) 
-      addtolist(have, funcptr->name, funcptr->tableunit, "<piecewise linear>", 
-		&list, &listsize, &maxnamelen, &count);
+      addtolist(have, searchstring, funcptr->name, funcptr->tableunit, 
+                "<piecewise linear>", &list, &listsize, &maxnamelen, &count, 
+                searchtype);
     else
-      addtolist(have, funcptr->name, funcptr->inverse.dimen, "<nonlinear>", 
-		&list, &listsize, &maxnamelen, &count);
+      addtolist(have, searchstring, funcptr->name, funcptr->inverse.dimen, 
+                "<nonlinear>", &list, &listsize, &maxnamelen, &count, 
+                searchtype);
   }
   qsort(list, count, sizeof(struct namedef), compnd);
 
@@ -2086,6 +2111,8 @@ usage()
     -e, --exponential   exponential format output\n\
     -f, --file          specify a units data file (-f '' loads default file)\n\
     -m, --minus         make - into a subtraction operator (default)\n\
+        --oldstar       use old '*' precedence, higher than '/'\n\
+        --newstar       use new '*' precedence, equal to '/'\n\
     -o, --output-format specify printf numeric output format\n\
     -p, --product       make - into a product operator\n\
     -q, --quiet         supress prompting\n\
@@ -2257,7 +2284,7 @@ findunitsfile()
 
     env = getenv("PATH");
     if (env) {
-      filename = mymalloc(strlen(env)+strlen(UNITSFILE)+2,"(readunits)");
+      filename = mymalloc(strlen(env)+strlen(UNITSFILE)+2,"(findunitsfile)");
 
       if (strchr(env, ';'))	/* MS-DOS */
 	strcpy(separator, ";");
@@ -2265,7 +2292,7 @@ findunitsfile()
 	strcpy(separator, ":");
       direc = strtok(env, separator);
       while (direc) {
-	strcpy(filename, "");
+	strcpy(filename, "");      /* Clear string for next iteration */
 	strcat(filename, direc);
 	strcat(filename, "/");
 	strcat(filename, UNITSFILE);
@@ -2283,7 +2310,30 @@ findunitsfile()
     fclose(testfile);
   return file;
 }
-   
+
+
+char *
+personalunitsfile()
+{
+  FILE *testfile;
+  char *homedir, *filename;
+
+  homedir = getenv("HOME");
+  if (!homedir)
+    return 0;
+  filename = mymalloc(strlen(homedir)+strlen(HOMEUNITSFILE)+2,"(personalunitsfile)");
+  strcpy(filename,homedir);
+  strcat(filename,"/");
+  strcat(filename,HOMEUNITSFILE);
+  testfile = fopen(filename, "rt");
+  if (testfile){
+    fclose(testfile);
+    return filename;
+  }
+  free(filename);
+  return 0;
+}
+
 
 char *shortoptions = "Vvqechstf:o:mp1";
 
@@ -2305,6 +2355,8 @@ struct option longoptions[] = {
   {"minus", no_argument, &minusminus, 1},
   {"product", no_argument, &minusminus, 0},
   {"one-line", no_argument, &oneline, 1},
+  {"oldstar", no_argument, &oldstar, 1},
+  {"newstar", no_argument, &oldstar, 0},
   {0,0,0,0} };
 
 /* Process the args.  Returns 1 if interactive mode is desired, and 0
@@ -2478,7 +2530,21 @@ ishelpquery(char *str, struct unittype *have)
   
   str=removepadding(str);
   if (have && !strcmp(str, UNITMATCH)){
-    tryallunits(have);
+    tryallunits(have,0);
+    return 1;
+  }
+  if (!strncmp(SEARCHCOMMAND,str,strlen(SEARCHCOMMAND))){
+    str+=strlen(SEARCHCOMMAND);
+    if (!strchr(WHITE,*str))
+      return 0;
+    str = removepadding(str);
+    if (strlen(str)==0){
+      printf("\n\
+Type 'search text' to see a list of all unit names \n\
+containing 'text' as a substring\n\n");
+      return 1;
+    }
+    tryallunits(0,str);
     return 1;
   }
   if (!strncmp(HELPCOMMAND,str,strlen(HELPCOMMAND))){
@@ -2506,9 +2572,9 @@ To quit from units type ^%s.\n\
 At the '%s' prompt press return to see the definition of the unit you\n\
 entered above or '%s' to get a list of conformable units. \n\
 \n\
-At either prompt you can type 'help' to see this message or 'help myunit' to \n\
-browse the units database and read the comments relating to myunit or see\n\
-other units related to myunit. \n\n",
+At either prompt you type 'help myunit' to browse the units database and\n\
+read the comments relating to myunit or see other units related to myunit.\n\
+Typing 'search text' will show units whose names contain 'text'.\n\n",
 	     queryhave, 
              querywant, 
              queryhave, queryhave, queryhave,
@@ -2575,10 +2641,15 @@ main(int argc, char **argv)
      deftext = "";
 
    if (!unitsfiles[0]){
-     unitsfiles[0] = getenv("UNITSFILE");
-     if (!unitsfiles[0])
-       unitsfiles[0] = findunitsfile();
-     unitsfiles[1] = 0;
+     unitsfiles[0] = personalunitsfile();
+     if (unitsfiles[0])
+       unitcount=1;
+     else
+       unitcount=0;
+     unitsfiles[unitcount] = getenv("UNITSFILE");
+     if (!unitsfiles[unitcount])
+       unitsfiles[unitcount] = findunitsfile();
+     unitsfiles[unitcount+1] = 0;
    }
 
    mylocale = getenv("LOCALE");
